import re
from copy import copy
from operator import itemgetter
from threading import Thread
from typing import Optional, List, Any

from PyQt6.QtCore import pyqtSlot, pyqtSignal, Qt, QSize
from PyQt6.QtGui import QMovie, QPixmap, QIcon
from PyQt6.QtWidgets import QWidget, QMessageBox
from crawlMp.results import Results

from crawlMpGui.templates.resultsWidgetTpl import Ui_Form
from crawlMpGui.widgets.results_view import ResultsViewModel


class ResultsWidget(QWidget, Ui_Form):
    """
    Base widget to show CrawlMp search Results.
    Results are paged into page_sized chunks.
    User can sort hits by column and apply multiple filter based on column type.
    """
    sig_update_results = pyqtSignal(Results)
    sig_sorting_done = pyqtSignal()
    page_size = 200
    current_page = 0
    pages_count = 0
    loading = False

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self.results: Optional[Results] = None
        self.results_copy: Optional[Results] = None
        # Preload resources
        self.done_icon = QPixmap("icons:done.png").scaledToHeight(24)
        self.loading_movie = QMovie("icons:loading.gif")
        self.loading_movie.setScaledSize(QSize(24, 24))
        self.loading_movie.start()

        self.sig_update_results.connect(self.set_results)
        self.sig_sorting_done.connect(self.hits_sorted)
        self.control_widgets = (
            self.input_current_page, self.input_category, self.input_filler, self.button_filter_apply,
            self.button_filter_reset)
        self.set_loading()

    @property
    def page_hits(self) -> List[Any]:
        page_start = self.current_page * self.page_size
        page_end = page_start + self.page_size
        return self.results.hits[page_start:page_end]

    def _sorting(self, order: Qt.SortOrder, column_index: int) -> None:
        """
        Sort Results by column in Ascending or Descending order.
        :param Qt.SortOrder order: order
        :param int column_index: column index
        :return: None
        """
        self.results.hits.sort(key=itemgetter(column_index), reverse=order == Qt.SortOrder.DescendingOrder)
        self.sig_sorting_done.emit()

    def _pattern_filtering(self, pattern: re.Pattern, column_index: int) -> None:
        """
        Filter string values by regular expression pattern.
        :param re.Pattern pattern: Filtering pattern
        :param int column_index: column index
        :return: None
        """
        if self.results_copy is None:
            self.results_copy = copy(self.results)
        self.results.hits = [s for s in self.results.hits if pattern.search(s[column_index])]
        self.sig_sorting_done.emit()

    def _eval_filtering(self, expression: str, column_index: int) -> None:
        """
        Filter numerical values by expression evaluation
        :param str expression: right side of expression
        :param int column_index: column index
        :return: None
        """
        if self.results_copy is None:
            self.results_copy = copy(self.results)
        self.results.hits = [s for s in self.results.hits if eval(f"{s[column_index]}{expression}")]
        self.sig_sorting_done.emit()

    def _update_enabled_states(self) -> None:
        """
        Update enabled state of elements
        :return: None
        """
        self.button_first_page.setDisabled(self.current_page == 0 or self.loading)
        self.button_previous_page.setDisabled(self.current_page == 0 or self.loading)
        self.button_last_page.setDisabled(self.current_page == self.pages_count or self.loading)
        self.button_next_page.setDisabled(self.current_page == self.pages_count or self.loading)
        for widget in self.control_widgets:
            widget.setDisabled(self.loading)
        if self.results_copy is None:
            self.button_filter_reset.setDisabled(True)

    @pyqtSlot(Results)
    def set_results(self, results: Results) -> None:
        """
        Set new Results
        :param Results results: New crawl Results
        :return: None
        """
        self.current_page = 0
        self.results = results
        self.pages_count = int(len(self.results.hits) / self.page_size)
        self.input_current_page.setMaximum(self.pages_count + 1)
        current_selected = self.input_category.currentText()
        self.input_category.blockSignals(True)
        self.input_category.clear()
        self.input_category.addItems((item[0].value for item in results.hits_header))
        self.input_category.setCurrentText(current_selected)
        self.input_category.blockSignals(False)
        self.input_current_page.setSuffix(f" / {self.pages_count + 1}")
        self.set_idle()
        self.change_current_page()

    def change_current_page(self) -> None:
        """
        Current page has been changed.
        Set view model and update controls enabled state.
        :return: None
        """
        count_offset = self.current_page * self.page_size
        model = ResultsViewModel(self.page_hits, self.results.hits_header, count_offset)
        model.sig_start_sorting.connect(self.sort_hits)
        self.input_current_page.setValue(self.current_page + 1)
        self.view_results.setModel(model)
        self._update_enabled_states()

    @pyqtSlot(int, Qt.SortOrder)
    def sort_hits(self, column_index: int, order: Qt.SortOrder) -> None:
        """
        Sort hits by column in Ascending or Descending order
        :param column_index:
        :param order:
        :return:
        """
        self.set_loading()
        sorting_thread = Thread(target=self._sorting, args=(order, column_index))
        sorting_thread.start()

    @pyqtSlot()
    def hits_sorted(self) -> None:
        """
        Sorting has been finished
        :return: None
        """
        self.set_results(self.results)
        self.change_current_page()

    @pyqtSlot(int)
    def set_page(self, number) -> None:
        """
        Show page number
        :param int number: page number
        :return: None
        """
        self.current_page = number - 1
        self.change_current_page()

    @pyqtSlot()
    def next_page(self) -> None:
        """
        Show next results page
        :return: None
        """
        self.current_page += 1
        self.change_current_page()

    @pyqtSlot()
    def previous_page(self) -> None:
        """
        Show previous results page
        :return: None
        """
        self.current_page -= 1
        self.change_current_page()

    @pyqtSlot()
    def first_page(self) -> None:
        """
        Show first results page
        :return: None
        """
        self.current_page = 0
        self.change_current_page()

    @pyqtSlot()
    def last_page(self) -> None:
        """
        Show last results page
        :return: None
        """
        self.current_page = self.pages_count
        self.change_current_page()

    @pyqtSlot()
    def set_loading(self) -> None:
        """
        Searching or sorting has been executed, Loading state.
        :return: None
        """
        self.loading = True
        self.label_status.setMovie(self.loading_movie)
        self.label_status.setToolTip("Loading")
        self.setDisabled(True)
        self._update_enabled_states()

    @pyqtSlot()
    def set_idle(self) -> None:
        """
        Searching or sorting is finished, Idle state.
        :return: None
        """
        self.loading = False
        self.label_status.setPixmap(self.done_icon)
        self.label_status.setToolTip("Idle")
        self.setDisabled(False)
        self._update_enabled_states()

    @pyqtSlot(int)
    def column_filter_changed(self, column_index: int) -> None:
        """
        Column filter value was changed by the user.
        Clear input_filter and set placeholder for expression filter.
        :param column_index:
        :return: None
        """
        header_name, header_type, header_unit = self.results.hits_header[column_index]
        self.input_filler.clear()
        if header_type in (float, int):
            self.input_filler.setPlaceholderText("Filter by value comparison (>, <, !=, ==, <=, >=)")
        else:
            self.input_filler.setPlaceholderText("Filter by regexp")

    @pyqtSlot()
    def filter_results(self) -> None:
        """
        Start filtering results.
        Check header type and set filter based on it.
        :return: None
        """
        header_index = self.input_category.currentIndex()
        header_name, header_type, header_unit = self.results.hits_header[header_index]
        filter_value = self.input_filler.text()
        if self.loading:
            QMessageBox.critical(self, "Error", "Loading is in progress!")
            return
        if filter_value == "":
            QMessageBox.critical(self, "Error", "Filter input is empty!")
            return

        if header_type in (float, int):
            if not re.match("^\s?([<>!][=]?|==)\s?[+-]?([0-9]*[.])?[0-9]+$", filter_value):
                QMessageBox.critical(self, "Error", "Wrong filter pattern!")
                return
            else:
                filter_method = self._eval_filtering
        else:
            filter_method, filter_value = self._pattern_filtering, re.compile(filter_value)

        self.set_loading()
        Thread(target=filter_method, args=(filter_value, header_index)).start()

    @pyqtSlot()
    def filter_reset(self) -> None:
        """
        Reset all filters
        :return: None
        """
        if self.results_copy is not None:
            self.set_results(self.results_copy)
            self.results_copy = None
            self._update_enabled_states()



# Form implementation generated from reading ui file 'resultsWidgetTpl.ui'
#
# Created by: PyQt6 UI code generator 6.2.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(837, 324)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.lyt_paging = QtWidgets.QHBoxLayout()
        self.lyt_paging.setObjectName("lyt_paging")
        self.label_status = QtWidgets.QLabel(Form)
        self.label_status.setMinimumSize(QtCore.QSize(24, 24))
        self.label_status.setText("")
        self.label_status.setPixmap(QtGui.QPixmap("icons:loading.gif"))
        self.label_status.setObjectName("label_status")
        self.lyt_paging.addWidget(self.label_status)
        self.button_first_page = QtWidgets.QPushButton(Form)
        self.button_first_page.setMaximumSize(QtCore.QSize(30, 16777215))
        self.button_first_page.setObjectName("button_first_page")
        self.lyt_paging.addWidget(self.button_first_page)
        self.button_previous_page = QtWidgets.QPushButton(Form)
        self.button_previous_page.setMaximumSize(QtCore.QSize(30, 16777215))
        self.button_previous_page.setObjectName("button_previous_page")
        self.lyt_paging.addWidget(self.button_previous_page)
        self.input_current_page = QtWidgets.QSpinBox(Form)
        self.input_current_page.setMinimumSize(QtCore.QSize(75, 0))
        self.input_current_page.setMinimum(1)
        self.input_current_page.setMaximum(9999999)
        self.input_current_page.setObjectName("input_current_page")
        self.lyt_paging.addWidget(self.input_current_page)
        self.button_next_page = QtWidgets.QPushButton(Form)
        self.button_next_page.setMaximumSize(QtCore.QSize(30, 16777215))
        self.button_next_page.setObjectName("button_next_page")
        self.lyt_paging.addWidget(self.button_next_page)
        self.button_last_page = QtWidgets.QPushButton(Form)
        self.button_last_page.setMaximumSize(QtCore.QSize(30, 16777215))
        self.button_last_page.setObjectName("button_last_page")
        self.lyt_paging.addWidget(self.button_last_page)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        self.lyt_paging.addItem(spacerItem)
        self.input_category = QtWidgets.QComboBox(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.input_category.sizePolicy().hasHeightForWidth())
        self.input_category.setSizePolicy(sizePolicy)
        self.input_category.setMinimumSize(QtCore.QSize(150, 0))
        self.input_category.setLayoutDirection(QtCore.Qt.LayoutDirection.RightToLeft)
        self.input_category.setObjectName("input_category")
        self.lyt_paging.addWidget(self.input_category)
        self.input_filler = QtWidgets.QLineEdit(Form)
        self.input_filler.setMinimumSize(QtCore.QSize(200, 0))
        self.input_filler.setObjectName("input_filler")
        self.lyt_paging.addWidget(self.input_filler)
        self.button_filter_apply = QtWidgets.QPushButton(Form)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("icons:arrow-right.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.button_filter_apply.setIcon(icon)
        self.button_filter_apply.setObjectName("button_filter_apply")
        self.lyt_paging.addWidget(self.button_filter_apply)
        self.button_filter_reset = QtWidgets.QPushButton(Form)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap("icons:reset.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        self.button_filter_reset.setIcon(icon1)
        self.button_filter_reset.setObjectName("button_filter_reset")
        self.lyt_paging.addWidget(self.button_filter_reset)
        self.gridLayout.addLayout(self.lyt_paging, 0, 0, 1, 1)
        self.view_results = QtWidgets.QTableView(Form)
        self.view_results.setStyleSheet("")
        self.view_results.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.view_results.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ContiguousSelection)
        self.view_results.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.view_results.setTextElideMode(QtCore.Qt.TextElideMode.ElideNone)
        self.view_results.setSortingEnabled(True)
        self.view_results.setObjectName("view_results")
        self.view_results.horizontalHeader().setSortIndicatorShown(True)
        self.view_results.horizontalHeader().setStretchLastSection(True)
        self.gridLayout.addWidget(self.view_results, 1, 0, 1, 1)

        self.retranslateUi(Form)
        self.button_previous_page.clicked.connect(Form.previous_page) # type: ignore
        self.button_next_page.clicked.connect(Form.next_page) # type: ignore
        self.button_first_page.clicked.connect(Form.first_page) # type: ignore
        self.button_last_page.clicked.connect(Form.last_page) # type: ignore
        self.input_current_page.valueChanged['int'].connect(Form.set_page) # type: ignore
        self.input_category.currentIndexChanged['int'].connect(Form.column_filter_changed) # type: ignore
        self.button_filter_apply.clicked.connect(Form.filter_results) # type: ignore
        self.button_filter_reset.clicked.connect(Form.filter_reset) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.button_first_page.setToolTip(_translate("Form", "Jump to first page"))
        self.button_first_page.setText(_translate("Form", "|<"))
        self.button_previous_page.setToolTip(_translate("Form", "Jump to previous page"))
        self.button_previous_page.setText(_translate("Form", "<"))
        self.input_current_page.setToolTip(_translate("Form", "Current page / Total pages"))
        self.button_next_page.setToolTip(_translate("Form", "Jump to next page"))
        self.button_next_page.setText(_translate("Form", ">"))
        self.button_last_page.setToolTip(_translate("Form", "Jump to last page"))
        self.button_last_page.setText(_translate("Form", ">|"))
        self.input_category.setToolTip(_translate("Form", "Select column to filter on"))
        self.input_filler.setPlaceholderText(_translate("Form", "Filter by regexp"))
        self.button_filter_apply.setToolTip(_translate("Form", "Apply filter"))
        self.button_filter_apply.setText(_translate("Form", " Apply"))
        self.button_filter_reset.setToolTip(_translate("Form", "Reset filters"))
        self.button_filter_reset.setText(_translate("Form", " Reset"))



from PyQt6 import QtCore
import os

__version__ = "0.0.7"

import crawlMpGui

path = os.path.dirname(crawlMpGui.__file__)
print(path)
QtCore.QDir.addSearchPath('icons', os.path.join(path, 'resources'))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="crawlMpGui",
    version="0.0.7",
    license='MIT',
    author="Martin Domaracký",
    author_email="domarm@comat.sk",
    description="GUI for CrawlMp",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/domarm-comat/crawlMpGui",
    packages=setuptools.find_packages(),
    package_data={'crawlMpGui.resources': ['*.*'],
                  'crawlMpGui.templates': ['*.*']},
    scripts=['crawlMpGui/scripts/search_fs_mp_gui'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
    ],
    install_requires=[
        "crawlMp>=0.3.7",
        "PyQt6>=6.2.2"
    ],
    extras_requires={
    },
    python_requires='>=3.6',
)



from typing import Any, List

from PyQt6.QtCore import QAbstractTableModel, QModelIndex, Qt, pyqtSignal
from PyQt6.QtGui import QColor
from crawlMp.enums import Header_ref


class ResultsViewModel(QAbstractTableModel):
    sig_start_sorting = pyqtSignal(int, Qt.SortOrder)

    def __init__(self, hits: List[Any], header: List[Header_ref], count_offset: int = 0) -> None:
        self.hits = hits
        self.header = header
        self.count_offset = count_offset
        self.row_color_1 = QColor("#ffffff")
        self.row_color_2 = QColor("#d7efe0")
        super().__init__()

    def rowCount(self, parent: QModelIndex = ...) -> int:
        return len(self.hits)

    def columnCount(self, parent: QModelIndex = ...) -> int:
        return len(self.header)

    def data(self, index: QModelIndex, role: int = ...) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            return f" {self.hits[index.row()][index.column()]} "
        elif role == Qt.ItemDataRole.BackgroundRole:
            return self.row_color_1 if index.row() % 2 else self.row_color_2

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = ...) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                unit = f" ({self.header[section][2]})" if self.header[section][2] is not None else ""
                return self.header[section][0].value + unit
            elif orientation == Qt.Orientation.Vertical:
                return f" {self.count_offset + section + 1} "

    def sort(self, column: int, order: Qt.SortOrder = ...) -> None:
        self.sig_start_sorting.emit(column, order)
